/* Define a target vector form ARM AIF format.  This is a skeletal BFD
   that only implements enough functions to be able to link a.out files to 
   form an AIF executable image.
   Copyright (C) 1990, 1991, 1992, 1993, 1994 Free Software Foundation, Inc.

This file is part of BFD, the Binary File Descriptor library.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "bfd.h"

#ifdef BFD64
#define unsigned32 unsigned int
#define int32 int
#define unsigned16 unsigned short
#else
#define unsigned32 unsigned long
#define int32 int
#define unsigned16 unsigned short
#endif

/* Only for a big endian target... */
/* #define BIG_ENDIAN_TARGET */

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "arm/aif.h"
#include "arm/asdfmt.h"


#ifdef BFD64
#define SECTION_ITEM_SIZE (sizeof(struct ItemSection) - 12)
#else
#define SECTION_ITEM_SIZE (sizeof(struct ItemSection))
#endif

#define READONLY_TARGET_INDEX 1
#define READWRITE_TARGET_INDEX 2
#define ZERO_INIT_TARGET_INDEX 3

/* AIF function prototypes */
void aif_swap_header(struct aif_hdr *);
void aif_swap_itemsection(struct ItemSection *);

bfd_target *dummy_check_format (bfd *junk);
bfd_target *check_aif_format (bfd *junk);
boolean bfd_set_format_to_object (bfd *junk);
boolean bfd_set_format_to_something_wierd (bfd *junk);
boolean bfd_write_object_contents (bfd *junk);
boolean bfd_write_wierd_format_contents (bfd *junk);

boolean AIF_close_and_cleanup (bfd *junk);
boolean AIF_bfd_free_cached_info (bfd *junk);
boolean AIF_new_section_hook (bfd *junk, sec_ptr rubbish);
boolean AIF_get_section_contents (bfd *junk, sec_ptr rubbish,
				  PTR useless, bfd_size_type wibble);

boolean AIF_bfd_copy_private_bfd_data (bfd *junk, bfd *rubbish);
boolean AIF_bfd_copy_private_section_data (bfd *junk, sec_ptr rubbish,
					   bfd *useless, sec_ptr wibble);

#define AIF_core_entry_point 0
#define AIF_core_file_failing_command AIF_core_entry_point
#define AIF_core_file_failing_signal AIF_core_entry_point
#define AIF_core_file_matches_executable_p AIF_core_entry_point

#define AIF_archive_entry_point 0
#define AIF_slurp_armap AIF_archive_entry_point
#define AIF_slurp_extended_name_table AIF_archive_entry_point
#define AIF_truncate_arname AIF_archive_entry_point
#define AIF_write_armap AIF_archive_entry_point
#define AIF_openr_next_archived_file AIF_archive_entry_point
#define AIF_generic_stat_arch_elt AIF_archive_entry_point

long AIF_get_symtab_upper_bound (bfd *junk);
long AIF_get_symtab (bfd *junk, struct symbol_cache_entry **rubbish);
asymbol *AIF_make_empty_symbol (bfd *junk);
void AIF_print_symbol (bfd *junk, PTR rubbish, struct symbol_cache_entry useless,
			   bfd_print_symbol_type wibble);
void AIF_get_symbol_info (bfd *junk, struct symbol_cache_entry rubbish,
			  symbol_info *useless);
alent *AIF_get_lineno (bfd *junk, struct symbol_cache_entry *useless);
boolean AIF_find_nearest_line (bfd *junk, struct sec *useless, struct symbol_cache_entry **rubbish,
			       bfd_vma wibble, CONST char **flibble, CONST char **dibble,
			       unsigned int *dong);
boolean AIF_bfd_is_local_label (bfd *junk, asymbol *useless);
asymbol *AIF_bfd_make_debug_symbol (bfd *junk, void *useless, unsigned long rubbish);


long AIF_get_reloc_upper_bound (bfd *junk, sec_ptr *rubbish);
long AIF_canonicalize_reloc (bfd *junk, sec_ptr rubbish, arelent **useless,
				 struct symbol_cache_entry **wibble);
CONST struct reloc_howto_struct *AIF_bfd_reloc_type_lookup(bfd *junk,
							  bfd_reloc_code_real_type rubbish);

boolean AIF_set_arch_mach (bfd *junk, enum bfd_architecture rubbish, unsigned long useless);
boolean AIF_set_section_contents (bfd *junk, sec_ptr rubbish, PTR useless,
				  file_ptr wibble, bfd_size_type flibble);

int AIF_sizeof_headers (bfd *junk, boolean rubbish);
bfd_byte *AIF_bfd_get_relocated_section_contents (bfd *junk,
						  struct bfd_link_info *rubbish,
						  struct bfd_link_order *useless,
						  bfd_byte *wibble, boolean flibble,
						  struct symbol_cache_entry **dibble);
boolean AIF_bfd_relax_section (bfd *junk, struct sec *rubbish,
			       struct bfd_link_info *useless, boolean *wibble);
struct bfd_link_hash_table *AIF_bfd_link_hash_table_create (bfd *junk);
boolean AIF_bfd_link_add_symbols (bfd *junk, struct bfd_link_info *rubbish);
boolean AIF_bfd_final_link (bfd *junk, struct bfd_link_info *rubbish);


#define dynamic_entry_point 0
#define AIF_get_dynamic_symtab_upper_bound dynamic_entry_point
#define AIF_canonicalize_dynamic_symtab dynamic_entry_point
#define AIF_get_dynamic_reloc_upper_bound dynamic_entry_point
#define AIF_canonicalize_dynamic_reloc dynamic_entry_point


char *AIF_backend_data = "Cheese";

/* The structure defined in the tdata union of a bfd */
/* Contains various snippets... */
struct aif_data_struct {
  char *today_is_a_friday;
  struct aif_hdr *header;
  struct ItemSection *debug;
  asection *read_only;
  asection *read_write;
  asection *zero_init;
  int offset_debug;
};
#define AIF_HEADER(x) ((((x)->tdata).aif_data)->header)
#define AIF_DEBUG(x) ((((x)->tdata).aif_data)->debug)
#define AIF_READ_ONLY(x) ((((x)->tdata).aif_data)->read_only)
#define AIF_READ_WRITE(x) ((((x)->tdata).aif_data)->read_write)
#define AIF_ZERO_INIT(x) ((((x)->tdata).aif_data)->zero_init)
#define AIF_OFFSET_DEBUG(x) ((((x)->tdata).aif_data)->offset_debug)

bfd_target aif_vec = {
  "aif",
  bfd_target_aif_flavour,
#ifdef BIG_ENDIAN_TARGET
  true,     /* data area byte sex */
  true,     /* header area byte sex */
#else
  false,
  false,
#endif
  (EXEC_P|WP_TEXT|HAS_SYMS|HAS_DEBUG),     /* The BARE necessities */
  (SEC_HAS_CONTENTS|SEC_ALLOC|SEC_LOAD|SEC_RELOC),
  0,     /* Hmmm... these four are a bit wierd especially */
  0,     /* the last one- minimum section alignment- 3!? */
  15,    /* Copied from aout-target.h */
  3,
#ifdef BIG_ENDIAN_TARGET
  bfd_getb64, bfd_getb_signed_64, bfd_putb64,
     bfd_getb32, bfd_getb_signed_32, bfd_putb32,
     bfd_getb16, bfd_getb_signed_16, bfd_putb16, /* data */
  bfd_getb64, bfd_getb_signed_64, bfd_putb64,
     bfd_getb32, bfd_getb_signed_32, bfd_putb32,
     bfd_getb16, bfd_getb_signed_16, bfd_putb16, /* hdrs */
#else
  bfd_getl64, bfd_getl_signed_64, bfd_putl64,
     bfd_getl32, bfd_getl_signed_32, bfd_putl32,
     bfd_getl16, bfd_getl_signed_16, bfd_putl16, /* data */
  bfd_getl64, bfd_getl_signed_64, bfd_putl64,
     bfd_getl32, bfd_getl_signed_32, bfd_putl32,
     bfd_getl16, bfd_getl_signed_16, bfd_putl16, /* hdrs */
#endif
  {dummy_check_format, check_aif_format,
     dummy_check_format, dummy_check_format},
  {bfd_set_format_to_something_wierd,
     bfd_set_format_to_object,
     bfd_set_format_to_something_wierd,
     bfd_set_format_to_something_wierd},
  {bfd_write_wierd_format_contents,
     bfd_write_object_contents,
     bfd_write_wierd_format_contents,
     bfd_write_wierd_format_contents},
  
  BFD_JUMP_TABLE_GENERIC (AIF),
  BFD_JUMP_TABLE_COPY (AIF),
  BFD_JUMP_TABLE_CORE (AIF),
  BFD_JUMP_TABLE_ARCHIVE (AIF),
  BFD_JUMP_TABLE_SYMBOLS (AIF),
  BFD_JUMP_TABLE_RELOCS (AIF),
  BFD_JUMP_TABLE_WRITE (AIF),
  BFD_JUMP_TABLE_LINK (AIF),
  BFD_JUMP_TABLE_DYNAMIC (AIF),
  
  &AIF_backend_data
};

void aif_swap_header(struct aif_hdr *header) {
  /* header is an array of unsigned 32's ... */
  unsigned32 *swap;
  int i;

  swap = (unsigned32 *) header;
  for (i = 0; i < sizeof(struct aif_hdr) / sizeof(unsigned32); i++)
    aif_vec.bfd_h_putx32(swap[i], swap + i);
}

void aif_swap_itemsection(struct ItemSection *sec) {
  unsigned32 *swap;
  int i;

  swap = (unsigned32 *) sec;
  aif_vec.bfd_putx32(swap[0], swap); /* Code and Length field */
  /* Miss out the four byte flags */
  for (i = 2; i < (sizeof (struct ItemSection)) / sizeof(unsigned32); i++)
    aif_vec.bfd_putx32(swap[i], swap + i);
}


bfd_target *dummy_check_format (bfd *junk) {
  return (bfd_target *) 0;
}

bfd_target *check_aif_format (bfd *abfd) {
  struct aif_hdr hdr;


  bfd_seek(abfd, 0, SEEK_SET);
  bfd_read(&hdr, 1, sizeof(struct aif_hdr), abfd);

  if ((hdr.compress_br == AIF_NOOP || (hdr.compress_br & 0xff000000 == 0xeb000000)) &&
      (hdr.reloc_br == AIF_NOOP || (hdr.reloc_br & 0xff000000 == 0xeb000000)) &&
      (hdr.zinit_br == AIF_NOOP || (hdr.zinit_br & 0xff000000 == 0xeb000000)) &&
      (hdr.entry_br == AIF_NOOP || (hdr.entry_br & 0xff000000 == 0xeb000000)))
    return &aif_vec;
  else
    return (bfd_target *) 0;
}


boolean bfd_set_format_to_object (bfd *abfd) {

  (abfd->tdata).aif_data = (struct aif_data_struct *) malloc(sizeof(struct aif_data_struct));
  AIF_HEADER(abfd) = (struct aif_hdr *) malloc(sizeof(struct aif_hdr));
  AIF_DEBUG(abfd) = (struct ItemSection *) malloc(sizeof(struct ItemSection));
  AIF_OFFSET_DEBUG(abfd) = sizeof(struct aif_hdr);
  ((abfd->tdata).aif_data)->today_is_a_friday="Whoopee! Today is a friday!\n";
  return true;
}

boolean bfd_set_format_to_something_wierd (bfd *junk) {

  return false;
}



boolean bfd_write_object_contents (bfd *abfd) {
  ItemSymbol aif_sym;
  unsigned32 asym, avalue;
  int i;
  int debug_size = 0;
  asymbol *sym;
  char **strings;
  int offset = 4;
  int padding;
  char space[4] = {0, 0, 0, 0};

/* Reserve enough pointers to the names of all the symbol names */
  strings = (char **) malloc (abfd->symcount * sizeof(char *));

/* Go to beginning of debug part of file */ 
  bfd_seek(abfd, AIF_OFFSET_DEBUG(abfd), SEEK_SET);
/* Leave enough room for the debug header */
  if (bfd_seek(abfd, SECTION_ITEM_SIZE, SEEK_CUR)) return false;


  for(i = 0; i < abfd->symcount; i++) {
    sym = abfd->outsymbols[i];
    asym = offset;

    if (sym->flags & BSF_GLOBAL || sym->flags & BSF_LOCAL) {
      if (sym->section->output_section == AIF_READ_ONLY(abfd))
	asym |= ASD_TEXTSYM;
      else if (sym->section->output_section == AIF_READ_WRITE(abfd))
	asym |= ASD_DATASYM;
      else if (sym->section->output_section == AIF_ZERO_INIT(abfd))
	asym |= ASD_ZINITSYM;
      else
	asym |= ASD_GLOBSYM;
    }

    avalue = sym->value;
    if (sym->flags == BSF_GLOBAL || sym->flags == BSF_LOCAL) {
      avalue += (sym->section)->output_offset;
      if (!sym->section->user_set_vma) avalue +=
	((sym->section)->output_section)->vma;
    }

    offset += strlen(sym->name) + 1;
    strings[i] = sym->name;

    aif_vec.bfd_putx32(asym, &aif_sym.sym);
    aif_vec.bfd_putx32(avalue, &aif_sym.value);
    bfd_write(&aif_sym, sizeof(ItemSymbol), 1, abfd);
  }
  
/* Now the string table */

/* pad to word boundary */
  padding = offset;
  offset = (offset + 3) & ~3;
  padding = offset - padding;

/* Put offset into aif_sym.value first to make sure it is 32 bits */
  aif_vec.bfd_putx32(offset, &avalue);

  bfd_write(&avalue, 4, 1, abfd);

  for(i = 0; i < abfd->symcount; i++)
    bfd_write(strings[i], 1, strlen(strings[i]) + 1, abfd);

  bfd_write(space, 1, padding, abfd);

  AIF_HEADER(abfd)->compress_br = AIF_NOOP;
  AIF_HEADER(abfd)->reloc_br = AIF_NOOP;
  AIF_HEADER(abfd)->zinit_br = AIF_NOOP;
  AIF_HEADER(abfd)->entry_br = AIF_BLAL | ((abfd->start_address - 0x8014) >> 2);
  AIF_HEADER(abfd)->exit_swi = OS_EXIT;
  AIF_HEADER(abfd)->rosize =
    AIF_READ_ONLY(abfd)->_raw_size + sizeof(struct aif_hdr);
  AIF_HEADER(abfd)->rwsize = AIF_READ_WRITE(abfd)->_raw_size;
  AIF_HEADER(abfd)->dbgsize =
    SECTION_ITEM_SIZE + 
      sizeof(ItemSymbol) * abfd->symcount + 
	offset;
  AIF_HEADER(abfd)->zinitsize = AIF_ZERO_INIT(abfd)->_raw_size;
  AIF_HEADER(abfd)->dbgtype = AIF_DBG_LL;
  AIF_HEADER(abfd)->imagebase = AIF_IMAGEBASE;
  AIF_HEADER(abfd)->workspace = AIF_ZERO_INIT(abfd)->_raw_size;
  AIF_HEADER(abfd)->address_mode = 32;  /* 32-bit addressing - no separate linking */
  AIF_HEADER(abfd)->data_base = AIF_IMAGEBASE + AIF_READ_WRITE(abfd)->filepos;
  AIF_HEADER(abfd)->debug_swi = AIF_NOOP;

  aif_swap_header(AIF_HEADER(abfd));  /* Fix byte sex */

  bfd_seek(abfd, 0, SEEK_SET);
  bfd_write(AIF_HEADER(abfd), 1, sizeof(struct aif_hdr), abfd);

/* Now write the debug header */
  bfd_seek(abfd, AIF_OFFSET_DEBUG(abfd), SEEK_SET);
  
  AIF_DEBUG(abfd)->id = (AIF_HEADER(abfd)->dbgsize << 16) | ITEMSECTION;
  AIF_DEBUG(abfd)->lang = LANG_NONE;
  AIF_DEBUG(abfd)->flags = 2; /* Only top level variables */
  AIF_DEBUG(abfd)->unused = 0;
  AIF_DEBUG(abfd)->asdversion = 2;

  AIF_DEBUG(abfd)->codestart = AIF_HEADER(abfd)->imagebase;
  AIF_DEBUG(abfd)->datastart = AIF_HEADER(abfd)->data_base;
  AIF_DEBUG(abfd)->codesize = AIF_HEADER(abfd)->rosize;
  AIF_DEBUG(abfd)->datasize = AIF_HEADER(abfd)->rwsize + AIF_HEADER(abfd)->zinitsize;
  AIF_DEBUG(abfd)->fileinfo.p = (ItemFileInfo *) 0;
  AIF_DEBUG(abfd)->debugsize = AIF_HEADER(abfd)->dbgsize;
  AIF_DEBUG(abfd)->n.nsyms = abfd->symcount;

  aif_swap_itemsection(AIF_DEBUG(abfd));

#ifndef BFD64
  bfd_write(AIF_DEBUG(abfd), 1, SECTION_ITEM_SIZE, abfd);
#else
/* There are unions of unsigned32's and pointers which are 64bit */
/* so a plain old bfd_write won't work! */
  bfd_write(AIF_DEBUG(abfd), 4, 6, abfd);  /* Upto the fileinfo union */
  avalue = 0;
  bfd_write(&avalue, 4, 1, abfd); /* fileinfo is always zero */
  avalue = AIF_DEBUG(abfd)->debugsize;
  bfd_write(&avalue, 4, 1, abfd);
  avalue = AIF_DEBUG(abfd)->n.nsyms;
  bfd_write(&avalue, 4, 1, abfd);
#endif

  free(strings);

  return true;
}

boolean bfd_write_wierd_format_contents (bfd *junk) {

  return false;
}







/* Generic entry points... */
boolean AIF_close_and_cleanup (bfd *junk) {

  return true;
}
boolean AIF_bfd_free_cached_info (bfd *junk) {

  return true;
}
boolean AIF_new_section_hook (bfd *abfd, sec_ptr newsect) {


  newsect->alignment_power = bfd_get_arch_info(abfd)->section_align_power;

  if (!strcmp(newsect->name, ".text") && !AIF_READ_ONLY(abfd)) {
    ((abfd->tdata).aif_data)->read_only = newsect;
    newsect->target_index = READONLY_TARGET_INDEX;
    newsect->filepos = sizeof(struct aif_hdr);
  }
  if (!strcmp(newsect->name, ".data") && !AIF_READ_WRITE(abfd)) {
    ((abfd->tdata).aif_data)->read_write = newsect;
    newsect->target_index = READWRITE_TARGET_INDEX;
    newsect->filepos = sizeof(struct aif_hdr);
  }
  if (!strcmp(newsect->name, ".bss") && !AIF_ZERO_INIT(abfd)) {
    ((abfd->tdata).aif_data)->zero_init = newsect;
    newsect->target_index = ZERO_INIT_TARGET_INDEX;
  }
  return true;
}
boolean AIF_get_section_contents (bfd *junk, sec_ptr rubbish,
				  PTR useless, bfd_size_type wibble) {

  return true;
}




/* Entry points to copy private data... */
boolean AIF_bfd_copy_private_bfd_data (bfd *junk, bfd *rubbish) {

  return true;
}
boolean AIF_bfd_copy_private_section_data (bfd *junk, sec_ptr rubbish,
					   bfd *useless, sec_ptr wibble) {

  return true;
}



/* Entry points used for symbol table */
long AIF_get_symtab_upper_bound (bfd *junk) {

  return 0;
}
long AIF_get_symtab (bfd *junk, struct symbol_cache_entry **rubbish) {

  return 0;
}


/* IMPLEMENTED IT!!! */
asymbol *AIF_make_empty_symbol (bfd *abfd) {
  asymbol *new;



  new = (asymbol *) bfd_zalloc (abfd, sizeof (asymbol)); /* An obstack memory allocation???? */
  if (!new)
    {
      bfd_set_error (bfd_error_no_memory);
      return NULL;
    }
  new->the_bfd = abfd;

  return new;

}





void AIF_print_symbol (bfd *junk, PTR rubbish, struct symbol_cache_entry useless,
		       bfd_print_symbol_type wibble) {

}
void AIF_get_symbol_info (bfd *junk, struct symbol_cache_entry rubbish,
			  symbol_info *useless) {

}
boolean AIF_bfd_is_local_label (bfd *junk, asymbol *useless) {

  return false;
}
alent *AIF_get_lineno (bfd *junk, struct symbol_cache_entry *useless) {

  return (alent *) 0;
}
boolean AIF_find_nearest_line (bfd *junk, struct sec *useless, struct symbol_cache_entry **rubbish,
       		       bfd_vma wibble, CONST char **flibble, CONST char **dibble,
			       unsigned int *dong) {

  return true;
}
asymbol *AIF_bfd_make_debug_symbol (bfd *junk, void *useless, unsigned long rubbish) {

  return (asymbol *) 0;
}







/* Routines for relocs.... */
long AIF_get_reloc_upper_bound (bfd *junk, sec_ptr *rubbish) {

  return 0;
}
long AIF_canonicalize_reloc (bfd *junk, sec_ptr rubbish, arelent **useless,
				 struct symbol_cache_entry **wibble) {

  return 0;
}
CONST struct reloc_howto_struct *AIF_bfd_reloc_type_lookup(bfd *junk,
							  bfd_reloc_code_real_type rubbish) {

  return (struct reloc_howto_struct *) 0;
}








/* Routines used when writing an object file */
/* IMPLEMENTED!!! */
boolean AIF_set_arch_mach (bfd *abfd, enum bfd_architecture arch, unsigned long mach) {

  if (arch != bfd_arch_arm) {

    return false;
  }
  else {

    if (!bfd_default_set_arch_mach(abfd, arch, mach)) return false;
    return true;
  }
}



boolean AIF_set_section_contents (bfd *abfd, sec_ptr section, PTR location,
				      file_ptr offset, bfd_size_type count) {


  if (section != AIF_ZERO_INIT(abfd))
    {
      if (bfd_seek(abfd, section->filepos + offset, SEEK_SET)) return false;
      if (count && (bfd_write(location, 1, count, abfd)==count)) {
	if (section == AIF_READ_ONLY(abfd)) {
	  AIF_READ_WRITE(abfd)->filepos += count;
	  AIF_OFFSET_DEBUG(abfd) += count;
	}
	else if (section == AIF_READ_WRITE(abfd))
	  AIF_OFFSET_DEBUG(abfd) += count;
	return true;
      }
      return false;
    }
  return true;
}









/* Routines used by the linker... */
int AIF_sizeof_headers (bfd *junk, boolean rubbish) {

  return 32;
}


bfd_byte *AIF_bfd_get_relocated_section_contents (bfd *abfd,
						  struct bfd_link_info *link_info,
						  struct bfd_link_order *link_order,
						  bfd_byte *data, boolean relocatable,
						  struct symbol_cache_entry **syms_in_cache) {


  
  return (bfd_generic_get_relocated_section_contents(abfd, link_info,
						     link_order, data,
						     relocatable, syms_in_cache));
}

boolean AIF_bfd_relax_section (bfd *junk, struct sec *rubbish,
			       struct bfd_link_info *useless, boolean *wibble) {

  return true;
}


/* IMPLEMENTED IT!!! */
struct bfd_link_hash_table *AIF_bfd_link_hash_table_create (bfd *abfd) {

  return(_bfd_generic_link_hash_table_create(abfd));
}



/* IMPLEMENTED THIS ASWELL!!! */
boolean AIF_bfd_link_add_symbols (bfd *abfd, struct bfd_link_info *link_info) {

  return(_bfd_generic_link_add_symbols(abfd, link_info));
}

boolean AIF_bfd_final_link (bfd *abfd, struct bfd_link_info *info) {

  return (_bfd_generic_final_link (abfd, info));
}


























